#include "COMIX/Run/Environment.H"

#include "MODEL/Main/Model_Base.H"
#include "PDF/Main/Structure_Function.H"
#include "PDF/Main/Intact.H"
#include "PDF/Main/PDF_Base.H"
#include "ATOOLS/Org/Data_Reader.H"
#include "ATOOLS/Org/Message.H"
#include "ATOOLS/Org/Shell_Tools.H"
#include "ATOOLS/Org/Data_Writer.H"
#include "ATOOLS/Org/Library_Loader.H"
#include "PHASIC++/Main/Phase_Space_Handler.H"
#include "PHASIC++/Selectors/Selector.H"
#include "PHASIC++/Process/ME_Generator_Base.H"
#include "ATOOLS/Org/MyStrStream.H"
#include "ATOOLS/Org/Run_Parameter.H"
#include "ATOOLS/Math/Random.H"

#include <sys/stat.h>

using namespace COMIX;
using namespace MODEL;
using namespace BEAM;
using namespace PDF;
using namespace ATOOLS;
using namespace std;

typedef void (*PDF_Init_Function)(const std::string &);
typedef void (*PDF_Exit_Function)();

#define PTP long unsigned int

Environment::Environment(int argc,char * argv[]) : 
  m_savestatus(false), p_model(NULL), p_beamspectra(NULL)
{
  m_path=std::string("./");
  m_file=std::string("Run.dat");

  std::vector<std::string> names(4);
  names[0]="Decaydata";
  names[1]="Run.dat";
  My_In_File::SetNoComplains(names);

  ExtractCommandLineParameters(argc, argv);
  SetFileNames();

  rpa.Init(m_path,m_file,argc,argv);
  LoadLibraries();
  ShowParameterSyntax();
  ran.InitExternal(m_path,m_file);

  exh->AddTerminatorObject(this);
}

void Environment::SetFileNames()
{
  p_dataread    = new Data_Reader(" ",";","!","=");
  p_dataread->AddComment("#");
  p_dataread->AddWordSeparator("\t");
  p_dataread->SetInputPath(m_path);
  p_dataread->SetInputFile(m_file);
  m_modeldat         = p_dataread->GetValue<string>("MODEL_DATA_FILE",string("Model.dat"));
  m_beamdat          = p_dataread->GetValue<string>("BEAM_DATA_FILE",string("Beam.dat"));
  m_isrdat           = p_dataread->GetValue<string>("ISR_DATA_FILE",string("ISR.dat"));
  m_medat            = p_dataread->GetValue<string>("ME_DATA_FILE",string("ME.dat"));
  std::string integrationdat=p_dataread->GetValue<string>
    ("INTEGRATION_DATA_FILE","Integration.dat");
  m_processesdat=p_dataread->GetValue<string>
    ("PROCESSFILE",string("Processes.dat"));
  m_selectordat=p_dataread->
    GetValue<string>("SELECTORFILE",string("Selector.dat"));

  std::string fname(m_file);
  if (fname.find("|")!=std::string::npos) 
    fname=fname.substr(0,fname.find("|"));
  Read_Write_Base cf(1,0," ",";","!","=");
  cf.SetAddCommandLine(false);
  cf.SetInputPath(m_path);
  cf.SetInputFile(fname+"|(beam){|}(beam)");
  if (cf.RereadInFile()) m_beamdat=fname+"|(beam){|}(beam)";
  cf.ClearFileBegin(); cf.ClearFileEnd();
  cf.SetInputFile(fname+"|(isr){|}(isr)");
  if (cf.RereadInFile()) m_isrdat=fname+"|(isr){|}(isr)";
  cf.ClearFileBegin(); cf.ClearFileEnd();
  cf.SetInputFile(fname+"|(model){|}(model)");
  if (cf.RereadInFile()) m_modeldat=fname+"|(model){|}(model)";
  cf.ClearFileBegin(); cf.ClearFileEnd();
  cf.SetInputFile(fname+"|(me){|}(me)");
  if (cf.RereadInFile()) m_medat=fname+"|(me){|}(me)";
  cf.ClearFileBegin(); cf.ClearFileEnd();
  cf.SetInputFile(fname+"|(processes){|}(processes)");
  if (cf.RereadInFile()) m_processesdat=fname+"|(processes){|}(processes)";
  cf.ClearFileBegin(); cf.ClearFileEnd();
  cf.SetInputFile(fname+"|(selector){|}(selector)");
  if (cf.RereadInFile()) m_selectordat=fname+"|(selector){|}(selector)";
  cf.ClearFileBegin(); cf.ClearFileEnd();
  cf.SetInputFile(fname+"|(integration){|}(integration)");
  if (cf.RereadInFile()) integrationdat=fname+"|(integration){|}(integration)";
  cf.ClearFileBegin(); cf.ClearFileEnd();

  rpa.gen.SetVariable("MODEL_DATA_FILE",m_modeldat);
  rpa.gen.SetVariable("ME_DATA_FILE",m_medat);
  rpa.gen.SetVariable("MODEL_DATA_FILE",m_modeldat);
  rpa.gen.SetVariable("INTEGRATION_DATA_FILE",integrationdat);
}


Environment::~Environment()
{
  if (m_savestatus) {
    msg_Error()<<METHOD<<"(): Status saved to '"
	       <<rpa.gen.Variable("SHERPA_STATUS_PATH")<<"'."<<std::endl;
    MakeDir(rpa.gen.Variable("SHERPA_STATUS_PATH"),493);
    exh->PrepareTerminate();
  }
  if (p_beamspectra)   { delete p_beamspectra;   p_beamspectra   = NULL; }
  if (p_model)         { delete p_model;         p_model         = NULL; }
  if (p_dataread)      { delete p_dataread;      p_dataread      = NULL; }
  std::set<ME_Handler*> deletedme;
  while (m_mehandlers.size()>0) {
    if (deletedme.find(m_mehandlers.begin()->second)==deletedme.end()) {
      deletedme.insert(m_mehandlers.begin()->second);
      delete m_mehandlers.begin()->second;
    }
    m_mehandlers.erase(m_mehandlers.begin());
  }
  while (m_isrhandlers.size()>0) {
    delete m_isrhandlers.begin()->second;
    m_isrhandlers.erase(m_isrhandlers.begin());
  }
  PHASIC::Phase_Space_Handler::DeleteInfo();
  exh->RemoveTerminatorObject(this);
  void *exit(s_loader->GetLibraryFunction(m_pdflib,"ExitPDFLib"));
  if (exit==NULL) THROW(fatal_error,"Cannot unload PDF library.");
  ((PDF_Exit_Function)(PTP)exit)();
}

void Environment::LoadLibraries() const
{
  Data_Reader read(" ",";","!","=");
  read.AddComment("#");
  read.SetInputFile(m_path+m_file);
  std::vector<std::string> ldadd;
  if (!read.VectorFromFile(ldadd,"SHERPA_LDADD")) return;
  for (size_t i(0);i<ldadd.size();++i) 
    if (!s_loader->LoadLibrary(ldadd[i])) 
      THROW(fatal_error,"Cannot load extra library.");
}

void Environment::ShowParameterSyntax()
{
  Data_Reader read(" ",";","!","=");
  int helpi(0);
  if (!read.ReadFromFile(helpi,"SHOW_ME_GENERATORS")) helpi=0;
  if (helpi>0) {
    msg->SetLevel(2);
    PHASIC::ME_Generator_Base::ShowSyntax(helpi);
    THROW(normal_exit,"Syntax shown.");
  }
  if (!read.ReadFromFile(helpi,"SHOW_SELECTOR_SYNTAX")) helpi=0;
  if (helpi>0) {
    msg->SetLevel(2);
    PHASIC::Selector_Base::ShowSyntax(helpi);
    THROW(normal_exit,"Syntax shown.");
  }
  if (!read.ReadFromFile(helpi,"SHOW_MODEL_SYNTAX")) helpi=0;
  if (helpi>0) {
    msg->SetLevel(2);
    MODEL::Model_Base::ShowSyntax(helpi);
    THROW(normal_exit,"Syntax shown.");
  }
}

std::string StripSectionTags(const std::string &name)
{
  if (name.find('|')!=std::string::npos)
    return name.substr(0,name.find('|'));
  return name;
}

void Environment::PrepareTerminate()
{
  std::string path(rpa.gen.Variable("SHERPA_STATUS_PATH")+"/");
  if (path=="/") return;
  CopyFile(m_path+StripSectionTags(m_file),path+StripSectionTags(m_file));
  CopyFile(m_path+StripSectionTags(m_modeldat),path+StripSectionTags(m_modeldat));
  CopyFile(m_path+StripSectionTags(m_beamdat),path+StripSectionTags(m_beamdat));
  CopyFile(m_path+StripSectionTags(m_isrdat),path+StripSectionTags(m_isrdat));
  CopyFile(m_path+StripSectionTags(m_medat),path+StripSectionTags(m_medat));
  CopyFile(m_path+StripSectionTags(m_selectordat),
	   path+StripSectionTags(m_selectordat));
  CopyFile(m_path+StripSectionTags(m_processesdat),
	   path+StripSectionTags(m_processesdat));
  CopyFile(m_path+StripSectionTags(rpa.gen.Variable("INTEGRATION_DATA_FILE")),
	   path+StripSectionTags(rpa.gen.Variable("INTEGRATION_DATA_FILE")));
  Data_Writer writer;
  writer.SetOutputFile(path+"cmd");
  writer.SetVectorType(vtc::vertical);
  writer.AddCommandLine("SHERPA_RUN_PATH = "+
			rpa.gen.Variable("SHERPA_RUN_PATH"));
  writer.AddCommandLine("SHERPA_CPP_PATH = "+
			rpa.gen.Variable("SHERPA_CPP_PATH"));
  writer.AddCommandLine("SHERPA_LIB_PATH = "+
			rpa.gen.Variable("SHERPA_LIB_PATH"));
  writer.VectorToFile(writer.CommandLine());
}

bool Environment::Initialize()
{
  bool okay = true;
  SetScaleFactors();
  okay = okay && InitializeTheModel();  
  okay = okay && InitializeTheBeams();
  okay = okay && InitializeThePDFs();
  ATOOLS::Integration_Info *info=PHASIC::Phase_Space_Handler::GetInfo();
  m_isrhandlers[isr::hard_process]->AssignKeys(info);
  if (m_isrhandlers.find(isr::hard_subprocess)!=m_isrhandlers.end()) {
    m_isrhandlers[isr::hard_subprocess]->AssignKeys(info);
  }
  if (!CheckBeamISRConsistency()) return 0.;
  okay = okay && InitializeTheMatrixElements();
  return okay;
}

bool Environment::CheckBeamISRConsistency()
{
  if (p_model->Name()==std::string("ADD")) {
    double ms = p_model->ScalarConstant("M_s");
    if (ms<rpa.gen.Ecms()) {
      msg_Error()<<"WARNING in Environment::CheckBeamISRConsistency :"<<std::endl
	       <<"   You are using the ADD model beyond its valid range ! "<<endl;
    }
  }

  double smin=0;
  double smax=sqr(rpa.gen.Ecms());
  smin = Max(smin,p_beamspectra->SprimeMin());
  smax = Min(smax,p_beamspectra->SprimeMax());
  if (m_isrhandlers[isr::hard_process]->On()) {
    smin = Max(smin,m_isrhandlers[isr::hard_process]->SprimeMin());
    smax = Min(smax,m_isrhandlers[isr::hard_process]->SprimeMax());
  }
  if (p_beamspectra->On()) {
    p_beamspectra->SetSprimeMin(smin);
  }
  string name=p_model->Name();
  if (name==std::string("ADD")) {
    double mcut2 = sqr(p_model->ScalarConstant("M_cut"));
    // if ISR & beam -> apply mcut on ISR only
    // if beam only  -> apply mcut on Beam
    smax = Min(smax,mcut2);
    for (size_t i=1;i<3;++i) {
      isr::id id=(isr::id)i;
      if (m_isrhandlers[id]->On()) {
	m_isrhandlers[id]->SetFixedSprimeMax(smax);
	m_isrhandlers[id]->SetFixedSprimeMin(smin);
      } 
      else if (p_beamspectra->On()) {
	p_beamspectra->SetSprimeMax(smax);
      }
    }
  }

  if (!(p_beamspectra->CheckConsistency(m_bunch_particles))) {
    msg_Error()<<"Error in Initialization of the Sherpa framework : "<<endl
	       <<"    Detected a mismatch of flavours from beams to bunches : "<<endl
	       <<"    "<<p_beamspectra->GetBeam(0)<<" -> "
	       <<m_isrhandlers[isr::hard_process]->Flav(0)<<" and "
	       <<p_beamspectra->GetBeam(1)<<" -> "
	       <<m_isrhandlers[isr::hard_process]->Flav(1)<<endl;
    return 0;
  }

  return 1;
}

bool Environment::InitializeTheModel()
{
  if (p_model) delete p_model;
  //determine and set scale for coupling initialization
  Data_Reader beamer(" ",";","!","=");
  beamer.AddComment("#");
  beamer.AddWordSeparator("\t");
  beamer.SetInputFile(m_path+m_beamdat);
  double beam1 = beamer.GetValue<double>("BEAM_ENERGY_1",0.0);
  double beam2 = beamer.GetValue<double>("BEAM_ENERGY_2",0.0);
  rpa.gen.SetCplScale(4.*beam1*beam2);
  Data_Reader read(" ",";","!","=");
  read.AddWordSeparator("\t");
  read.SetInputPath(m_path);
  read.SetInputFile(m_modeldat);
  std::string name;
  if (!read.ReadFromFile(name,"MODEL")) name="SM";
  p_model=Model_Base::Model_Getter_Function::
    GetObject(name,Model_Arguments(m_path,m_modeldat,true));
  if (p_model==NULL) THROW(not_implemented,"Model not implemented");
  p_model->InitializeInteractionModel();
  MODEL::s_model=p_model;
  return 1;
}


bool Environment::InitializeTheBeams()
{
  if (p_beamspectra) { delete p_beamspectra; p_beamspectra = NULL; }
  Data_Reader dataread(" ",";","!","=");
  dataread.AddComment("#");
  dataread.AddWordSeparator("\t");
  dataread.SetInputPath(m_path);
  dataread.SetInputFile(m_beamdat);
  p_beamspectra        = new Beam_Spectra_Handler(&dataread);
  msg_Info()<<"Initialized the beams "<<p_beamspectra->Type()<<endl;
  return 1;
}

bool Environment::InitializeThePDFs()
{
  Data_Reader dataread(" ",";","!","=");
  dataread.AddComment("#");
  dataread.AddWordSeparator("\t");
  dataread.SetInputPath(m_path);
  dataread.SetInputFile(m_isrdat);
#ifdef USING__LHAPDF
  std::string defaultlib("LHAPDFSherpa");
#else
  std::string defaultlib("CTEQ6Sherpa");
#endif  
  if (rpa.gen.Beam1().IsLepton() &&
      rpa.gen.Beam2().IsLepton()) defaultlib="PDFESherpa";
  m_pdflib=dataread.GetValue<std::string>("PDF_LIBRARY", defaultlib);
  void *init(s_loader->GetLibraryFunction(m_pdflib,"InitPDFLib"));
  if (init==NULL) THROW(fatal_error,"Cannot load PDF library.");
  std::string defset, defpath;
  if (m_pdflib=="LHAPDFSherpa") {
    defset="cteq6l.LHpdf";
    defpath="PDFSets";
  }
  else if (m_pdflib=="CTEQ6Sherpa") {
    defset="cteq6l";
    defpath="CTEQ6Grid";
  }
  else if (m_pdflib=="MRST04QEDSherpa") {
    defset="MRST04QED";
    defpath="MRST04Grid";
  }
  std::string grid_path=dataread.GetValue<string>("PDF_GRID_PATH",defpath);
  if (grid_path.length()==0 || grid_path[0]!='/')
    grid_path=rpa.gen.Variable("SHERPA_SHARE_PATH")+"/"+grid_path;
  ((PDF_Init_Function)(PTP)init)(grid_path);
  int helpi(0);
  if (!dataread.ReadFromFile(helpi,"SHOW_PDF_SETS")) helpi=0;
  if (helpi>0) {
    msg->SetLevel(2);
    PDF::PDF_Base::ShowSyntax(helpi);
    THROW(normal_exit,"Syntax shown.");
  }
  {
    isr::id id=(isr::id)1;
    if (m_isrhandlers.find(id)!=m_isrhandlers.end()) 
      delete m_isrhandlers[id]; 
    dataread.SetInputFile(m_isrdat);
    PDF_Base * pdfbase;
    ISR_Base ** isrbases = new ISR_Base*[2];
    double m_bunch_splimits[2];
    for (int j=0;j<2;++j) {
      int defaultflav(0);
      if (j==0) {
	defaultflav=rpa.gen.Beam1().IsAnti()? 
	  -rpa.gen.Beam1().Kfcode() : rpa.gen.Beam1().Kfcode();
      }
      else if (j==1) {
	defaultflav=rpa.gen.Beam2().IsAnti()? 
	  -rpa.gen.Beam2().Kfcode() : rpa.gen.Beam2().Kfcode();
      }
      int flav = dataread.GetValue<int>("BUNCH_"+ToString(j+1),defaultflav);
      m_bunch_particles[j] = Flavour((kf_code)abs(flav));
      if (flav<0) m_bunch_particles[j] = m_bunch_particles[j].Bar();
      std::string set = dataread.GetValue<std::string>("PDF_SET",defset);
      pdfbase = PDF_Base::PDF_Getter_Function::GetObject
	(set,PDF_Arguments(m_bunch_particles[j],grid_path,&dataread));
      if (m_bunch_particles[j].IsHadron() && pdfbase==NULL)
	THROW(critical_error,"PDF '"+set+"' does not exist in 'lib"+m_pdflib
	      +"' for "+ToString(m_bunch_particles[j])+" bunch.");
      msg_Info()<<"PDF set '"<<set<<"' loaded from 'lib"
		<<m_pdflib<<"'."<<std::endl;
      if (pdfbase==NULL) isrbases[j] = new Intact(m_bunch_particles[j]);     
      else isrbases[j] = new Structure_Function(pdfbase,m_bunch_particles[j]);
      ATOOLS::rpa.gen.SetBunch(m_bunch_particles[j],j);
    }
    m_bunch_splimits[0] = dataread.GetValue<double>("ISR_SMIN",1e-10);
    m_bunch_splimits[1] = dataread.GetValue<double>("ISR_SMAX",1.);
    double kplimits[2];
    kplimits[0] = dataread.GetValue<double>("ISR_KPMIN",m_bunch_splimits[0]);
    kplimits[1] = dataread.GetValue<double>("ISR_KPMAX",m_bunch_splimits[1]);
    m_isrhandlers[id] = new ISR_Handler(isrbases);
    m_isrhandlers[id]->SetBeam(p_beamspectra->GetBeam(0),0);
    m_isrhandlers[id]->SetBeam(p_beamspectra->GetBeam(1),1);
    m_isrhandlers[id]->Init(m_bunch_splimits,kplimits);
    msg_Info()<<"Initialized the ISR: "<<m_isrhandlers[id]->Type()<<endl;
    if (!(p_beamspectra->CheckConsistency(m_bunch_particles))) {
      msg_Error()<<"Error in Environment::InitializeThePDFs()"<<endl
		 <<"   Inconsistent ISR & Beam:"<<endl
		 <<"   Abort program."<<endl;
      abort();
    }
  }
  return 1;
}

bool Environment::InitializeTheMatrixElements()
{
  ME_Handler * me = NULL;
  me = new ME_Handler(m_path,m_medat,m_processesdat,m_selectordat);
  me->InitializeProcesses(p_model,p_beamspectra,m_isrhandlers[isr::hard_process]);
  MEHandlersMap::iterator it=m_mehandlers.find("SignalMEs");
  if (it!=m_mehandlers.end()) delete it->second;
  m_mehandlers["SignalMEs"]=me; 
  msg_Info()<<"Initialized the ME_Handler for the hard processes."
            <<endl;
  return 1;
}

ME_Handler * const Environment::GetMatrixElementHandler(std::string _key) { 
  MEHandlerIter pos = m_mehandlers.find(_key);
  if (pos!=m_mehandlers.end()) return pos->second;
  msg_Error()<<"Error in Environment::GetMatrixElementHandler("<<_key<<") :"
		     <<"   Key not found. Return Null pointer."<<endl;
  return NULL;
}


bool Environment::CalculateTheHardProcesses()
{
  ME_Handler * me = GetMatrixElementHandler(std::string("SignalMEs"));
  msg_Events()<<"=========================================================================="<<std::endl
              <<"Start calculating the hard cross sections. This may take some time.       "<<std::endl;
  int ok = me->CalculateTotalXSecs();
  if (ok) {
    msg_Events()<<"Calculating the hard cross sections has been successful.                  "<<std::endl
	     <<"=========================================================================="<<std::endl;
  }
  else {
    msg_Events()<<"Calculating the hard cross sections failed. Check this carefully.         "<<std::endl
	     <<"=========================================================================="<<std::endl;
  }
  return ok;
}

void Environment::SetScaleFactors() 
{
  Data_Reader dr(" ",";","!","=");
  dr.AddComment("#");
  dr.AddWordSeparator("\t");
  dr.SetInputPath(m_path);
  dr.SetInputFile(m_medat);
  double sf(dr.GetValue<double>("SCALE_FACTOR",1.));
  rpa.gen.SetVariable("FACTORIZATION_SCALE_FACTOR",
		      ToString(sf*dr.GetValue<double>("FACTORIZATION_SCALE_FACTOR",1.0)));
  rpa.gen.SetVariable("RENORMALIZATION_SCALE_FACTOR",
		      ToString(sf*dr.GetValue<double>("RENORMALIZATION_SCALE_FACTOR",1.0)));
  msg_Debugging()<<METHOD<<"(): Set scale factors {\n"
		 <<"  fac scale: "<<rpa.gen.Variable("FACTORIZATION_SCALE_FACTOR")<<"\n"
		 <<"  ren scale: "<<rpa.gen.Variable("RENORMALIZATION_SCALE_FACTOR")<<"\n}\n";
}

bool Environment::ExtractValArg
(std::vector<std::string> &args,std::vector<std::string>::iterator &it,
 const std::string &arg,const std::string &tag,const std::string &def) const
{
  if (it->find(arg)!=0) return false;
  std::string val=it->substr(2);
  if (def!="") {
    if (val!="") *it="-"+val;
    else it=args.erase(it);
    it=args.insert(it,tag+"="+def);
    return true;
  }
  if (val=="") {
    if (it+1!=args.end()) {
      it=args.erase(it);
      val=*it;
    }
  }
  if (val=="") {
    msg_Error()<<METHOD<<"(): No argument to '"
	       <<arg<<"'. Abort."<<std::endl;
    exit(1);
  }
  *it=tag+"="+val;
  return true;
}

int Environment::ExtractCommandLineParameters(int argc,char * argv[])
{
  std::string datpath;
  std::vector<std::string> helpsv(argc-1);
  for (int i(0);i<argc-1;++i) helpsv[i]=argv[i+1];
  for (std::vector<std::string>::iterator oit(helpsv.begin());
       oit!=helpsv.end();) {
    string par = *oit;
    string key,value;
    size_t equal=par.find("=");
    if (equal!=std::string::npos) {
      value = par.substr(equal+1);
      key   = par = par.substr(0,equal);
      if (key=="PATH") {
	if (value[value.length()-1]=='/') value.erase(value.length()-1,1);
	m_path=value;
        oit=helpsv.erase(oit);
      }
      else if (key=="RUNDATA") {
	m_file=value;
        oit=helpsv.erase(oit);
      }
      else if (key=="STATUS_PATH") {
	if (value[value.length()-1]!='/') value+=std::string("/");
	datpath=value;
        oit=helpsv.erase(oit);
      }
      else if (key=="SAVE_STATUS") {
	if (value[value.length()-1]!='/') value+=std::string("/");
	rpa.gen.SetVariable
	  ("SHERPA_STATUS_PATH",rpa.gen.Variable("SHERPA_RUN_PATH")+"/"+value);
	m_savestatus=true;
        oit=helpsv.erase(oit);
      }
      else {
	++oit;
      }
    }
    else if (ExtractValArg(helpsv,oit,"-f","RUNDATA"));
    else if (ExtractValArg(helpsv,oit,"-p","PATH"));
    else if (ExtractValArg(helpsv,oit,"-e","EVENTS"));
    else if (ExtractValArg(helpsv,oit,"-r","RESULT_DIRECTORY"));
    else if (ExtractValArg(helpsv,oit,"-m","ME_SIGNAL_GENERATOR"));
    else if (ExtractValArg(helpsv,oit,"-w","EVENT_GENERATION_MODE"));
    else if (ExtractValArg(helpsv,oit,"-F","FRAGMENTATION"));
    else if (ExtractValArg(helpsv,oit,"-a","ANALYSIS"));
    else if (ExtractValArg(helpsv,oit,"-g","GENERATE_RESULT_DIRECTORY","1"));
    else if (ExtractValArg(helpsv,oit,"-b","BATCH_MODE","0"));
    else if (ExtractValArg(helpsv,oit,"-O","OUTPUT"));
    else if (par=="--version" || par=="-v"){
      msg_Out()<<"Sherpa Version "<<SHERPA_VERSION<<"."<<SHERPA_SUBVERSION<<endl;
      exit(0);
    }
    else {
      if (par!="-h" && par!="--help")
	msg_Out()<<"Unrecognized option '"<<par<<"'.\n"<<endl;
      msg_Out()<<"Usage:\n"<<endl;
      msg_Out()<<"  Sherpa [options] [<parameter>=<value>] [<tag>:=<value>]\n"<<endl;
      msg_Out()<<"Options:\t-f <file>         read input from file <file>"<<endl;
      msg_Out()<<"\t\t-p <path>         read input from path <path>"<<endl;
      msg_Out()<<"\t\t-e <events>       set number of events <events>"<<endl;
      msg_Out()<<"\t\t-r <results>      set result directory <results>"<<endl;
      msg_Out()<<"\t\t-m <generator>    set me generator <generator>"<<endl;
      msg_Out()<<"\t\t-w <mode>         set event generation mode <mode>"<<endl;
      msg_Out()<<"\t\t-F <module>       set fragmentation module <module>"<<endl;
      msg_Out()<<"\t\t-a <analysis>     set analysis handler <analysis>"<<endl;
      msg_Out()<<"\t\t-O <level>        set output level <level>"<<endl;
      msg_Out()<<"\t\t-g                create result directory automatically"<<endl;
      msg_Out()<<"\t\t-b                run in non-batch mode"<<endl;
      msg_Out()<<"\t\t-v,--version      print the version number"<<endl;
      msg_Out()<<"\t\t-h,--help         print this help message\n"<<endl;
      exit(0);
    }
  }
  if (m_file.find("|")==std::string::npos) {
    Read_Write_Base cf(1,0," ",";","!","=");
    cf.SetInputPath(m_path);
    cf.SetInputFile(m_file+"|(run){|}(run)");
    if (cf.OpenInFile()) m_file+="|(run){|}(run)";
  }

  std::vector<std::string> helpsv2;
  // Add parameters from possible global.dat to command line
  Data_Reader dr(" ",";","!","=");
  dr.AddWordSeparator("\t");
  dr.AddComment("#");
  dr.SetInputPath(rpa.gen.Variable("HOME")+"/.sherpa/");
  dr.SetInputFile("global.dat");
  std::vector<std::vector<std::string> > helpsvv;
  if (dr.MatrixFromFile(helpsvv,"")) {
    msg_Out()<<METHOD<<"(): Reading parameters from '"
	     <<rpa.gen.Variable("HOME")<<"/.sherpa/global.dat'."<<std::endl;
    helpsv2.resize(helpsvv.size());
    for (size_t i(0);i<helpsvv.size();++i) {
      helpsv2[i]=helpsvv[i][0];
      for (size_t j(1);j<helpsvv[i].size();++j) helpsv2[i]+=" "+helpsvv[i][j];
    }
  }
  // Add parameters from Run.dat to command line
  // (this makes it possible to overwrite particle properties in Run.dat)
  dr.SetInputPath(m_path);
  dr.SetInputFile(m_file);
  dr.RereadInFile();
  if (dr.MatrixFromFile(helpsvv,"")) {
    size_t oldsize(helpsv2.size());
    helpsv2.resize(oldsize+helpsvv.size());
    for (size_t i(0);i<helpsvv.size();++i) {
      helpsv2[oldsize+i]=helpsvv[i][0];
      for (size_t j(1);j<helpsvv[i].size();++j)
	helpsv2[oldsize+i]+=" "+helpsvv[i][j];
    }
  }
  helpsv2.insert(helpsv2.end(),helpsv.begin(),helpsv.end());
  for (size_t i(0);i<helpsv2.size();++i) {
    string par = helpsv2[i];
    string key,value;
    size_t equal=Min(par.find("="),par.find(" "));
    if (equal!=std::string::npos) {
      value = par.substr(equal+1);
      key   = par = par.substr(0,equal);
      if (key[key.length()-1]==':') {
        key.erase(key.length()-1,1);
        Read_Write_Base::AddGlobalTag(key,value);
      }
      else {
        Read_Write_Base::AddCommandLine(key+" = "+value+"; ");
      }
    }
    else {
      Read_Write_Base::AddCommandLine(par+";");
    }
  }
  rpa.gen.SetVariable("RUN_DATA_FILE",m_file);

  if (datpath!="") m_path=datpath;
  std::vector<std::string> searchpaths;
  searchpaths.push_back(rpa.gen.Variable("SHERPA_RUN_PATH")+"/"+m_path);
  My_Out_File::SetSearchPaths(searchpaths);
  searchpaths.push_back(rpa.gen.Variable("SHERPA_DAT_PATH")+"/"+m_path);
  searchpaths.push_back(rpa.gen.Variable("SHERPA_DAT_PATH"));
  searchpaths.push_back(rpa.gen.Variable("SHERPA_SHARE_PATH")+"/"+m_path);
  searchpaths.push_back(rpa.gen.Variable("SHERPA_SHARE_PATH"));
  My_In_File::SetSearchPaths(searchpaths);
  rpa.gen.SetVariable("PATH_PIECE",m_path);
  m_path="";
  return 1;
}


